import base64
import gzip
import json
import os
import shutil
import unittest
import tempfile
from datetime import datetime, timedelta
from io import StringIO
from typing import Tuple, List
import sys
from breeze_email_reports.EmailProfileReport import main

TEST_REFERENCE_WRONG_NAME = '2023-09-16T23:00:24.211429.json'
TEST_REFERENCE_NAME = '2023-09-17T23:00:24.211429.json'
TEST_REPLAY_NAME = '2023-10-17T23:00:24.211429.json'

TEST_FILES_DIR = os.path.join(os.path.split(__file__)[0], 'test_files')

# Data from previous run
TEST_PREVIOUS_DATA = os.path.join(TEST_FILES_DIR, 'ReferenceData.json')

# Returned by Breeze for "current" data
CURRENT_DATA_BREEZE = os.path.join(TEST_FILES_DIR, 'CurrentDataBreeze.json')
# Output from command, processed CURRENT_DATA_BREEZE. (Though this is uncompressed.)
TEST_CURRENT_DATA = os.path.join(TEST_FILES_DIR, 'ExpectedCurrentData.json')

# Expected results in email
EXPECTED_TABLE = os.path.join(TEST_FILES_DIR, 'ExpectedHtml.txt')
EXPECTED_CSV = os.path.join(TEST_FILES_DIR, 'ExpectedCSV.csv')
EXPECTED_TEXT = os.path.join(TEST_FILES_DIR, 'ExpectedText.txt')

TO_ADDRESS = 'to@test.com'
BCC_ADDRESS = 'bcc1@bcc.com, bcc2@bcc.com'
CC_ADDRESS = 'cc@cc.com'


class MockBreezeAPI:
    def __init__(self, fields, profiles):
        self.fields = fields
        self.profiles = profiles

    def list_people(self, **kwargs):
        return self.profiles

    def get_profile_fields(self):
        return self.fields


class MockSender:
    def __init__(self):
        self.result = None

    def send_message(self, msg):
        self.result = msg


class TestSender(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.mock_sender = MockSender()
        self.saved_argv = sys.argv
        self.saved_stdout = None

    def runTest(self,
                report_format: str = 'text',
                do_gz: bool = True,
                do_reference: bool = True,
                previous_data: str = TEST_PREVIOUS_DATA,
                extra_params: List[str] = []) -> Tuple[str, int]:
        """
        Run BreezeProfile report. Return emitted event.
        :param report_format:
        :param do_gz:
        :param do_reference:
        :param previous_data:
        :param extra_params:
        :return:
        :raises: ParamException, IOException
        """

        with open(previous_data, 'r') as f:
            previous_data = f.read()

        test_file_name = os.path.join(self.test_dir.name, TEST_REFERENCE_NAME)
        if do_reference:
            if do_gz:
                with gzip.open(f'{test_file_name}.gz', 'w') as outfile:
                    outfile.write(bytes(previous_data, 'utf8'))
            else:
                with open(test_file_name, 'w') as f:
                    f.write(previous_data)
            with open(os.path.join(self.test_dir.name, TEST_REFERENCE_WRONG_NAME),
                      'w') as f:
                f.write('This is bad data')

        with open(CURRENT_DATA_BREEZE, 'r') as f:
            fields, profiles = json.load(f)

        api_mock = MockBreezeAPI(fields, profiles)

        sys.argv = ['test', '-f', 'from@test.com', '-t', TO_ADDRESS,
                    '-b', BCC_ADDRESS,
                    '--cc', CC_ADDRESS,
                    '--data', self.test_dir.name,
                    '--log_level=critical',
                    f'--logfile={os.path.join(self.test_dir.name, "test.log")}',
                    '--format', report_format] + extra_params

        ret = main(breeze_api=api_mock, email_sender=self.mock_sender)
        self.assertEqual(None, ret)
        # files = os.listdir(self.test_dir.name)
        # newfile = files[-1]
        # with open(os.path.join(self.test_dir.name, newfile), 'r') as f:
        #     expect_out = f.read()
        #
        # with open(TEST_CURRENT_DATA, 'w') as f:
        #     f.write(expect_out)

    def test_html(self):
        self.runTest('html',
                     extra_params=['--reference_data', TEST_PREVIOUS_DATA])

        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('to'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('bcc'))
        self.assertEqual(CC_ADDRESS, sent_message.get('cc'))

        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(str(pl0.get_payload()).
                        startswith('Breeze changes for Unknown thru'))
        self.assertEqual(pl1.get_content_subtype(), 'html')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TABLE+'.txt', 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TABLE, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_csv(self):
        self.runTest('csv', do_gz=True)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        # sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes since Sep 17 2023 11:00PM thru')))
        self.assertEqual(pl1.get_content_subtype(), 'csv')
        got_bytes = base64.b64decode(pl1.get_payload())
        got = got_bytes.decode('utf-8)')
        # with open(EXPECTED_CSV, 'w', newline='') as f:
        #     f.write(got)
        with open(EXPECTED_CSV, 'r', newline='') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_replay(self):
        shutil.copyfile(TEST_CURRENT_DATA,
                        os.path.join(self.test_dir.name, TEST_REPLAY_NAME))
        self.runTest('html', extra_params=['--replay'], do_gz=False)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertEqual(
            'Breeze changes for Sep 17 2023 11:00PM thru Oct 17 2023 11:00PM\n',
            str(pl0.get_payload()))
        # self.assertTrue(str(pl0.get_payload()).
        #                 startswith('Breeze changes for Unknown thru'))
        self.assertEqual(pl1.get_content_subtype(), 'html')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TABLE, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TABLE, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_no_change(self):
        self.runTest('csv', previous_data=TEST_CURRENT_DATA)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))
        payloads = sent_message.get_payload()
        self.assertEqual(len(payloads), 1)
        pl0 = payloads[0]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('No changes found between Sep 17 2023 11:00PM thru ')))

    def test_text(self):
        self.runTest('text')
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes for Sep 17 2023 11:00PM')))
        self.assertEqual(pl1.get_content_subtype(), 'plain')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TEXT, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TEXT, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_bad_params(self):
        # Exception if no "to"
        sys.argv = ['test', '-b', 'foo@bar']
        with self.assertRaises(SystemExit) as se:
            main(email_sender=self.mock_sender)
        self.assertEqual('--from=sender is required', se.exception.code)

    def test_no_to(self):
        sys.argv = ['test', '-f', 'from@foobar', ]
        with self.assertRaises(SystemExit) as se:
            main(email_sender=self.mock_sender)
        self.assertEqual('At least one of -t, -c, or -b is required', se.exception.code)

    def test_bad_data(self):
        sys.argv = ['test', '-f', 'from@foobar', '-t', 'to@foobar',
                    '--logfile=stdout',
                    '--log_level=critical',
                    '--data', '/no/such/dir']
        with self.assertRaises(SystemExit) as e:
            main(email_sender=self.mock_sender)
        self.assertEqual('Directory /no/such/dir doesn\'t exist',
                         e.exception.code)

    def test_no_previous(self):
        self.runTest(report_format='text', do_reference=False)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        # sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes for Unknown thru')))
        self.assertEqual(pl1.get_content_subtype(), 'plain')
        # got = str(pl1.get_payload())
        # print(got)
        # with open(EXPECTED_TEXT, 'w') as f:
        #     f.write(got)

    def test_retain(self):
        now = datetime.now()
        days = 7
        files = [datetime.isoformat(now - timedelta(days=days + d)) + '.json'
                 for d in [5, 1, -1, -2]]
        for f in files:
            shutil.copyfile(TEST_CURRENT_DATA, os.path.join(self.test_dir.name, f))

        self.runTest(report_format='text',
                     do_reference=False,
                     extra_params=['--retain_days', str(days)])
        retained_files = os.listdir(self.test_dir.name)
        self.assertEqual(3, len(retained_files))
        self.assertFalse(files[1] in retained_files)
        self.assertTrue(files[3] in retained_files)

    def test_list_directories(self):
        sys.argv = ['test', '--list_directories', '--log_level=critical']
        self.saved_stdout = sys.stdout
        capture_out = StringIO()
        sys.stdout = capture_out
        with self.assertRaises(SystemExit) as e:
            main(email_sender=self.mock_sender)
        self.assertEqual(e.exception.code, 0,
                         "List directories shouldn't cause error")
        sys.stdout = self.saved_stdout
        capture_out.seek(0)
        output = capture_out.read()
        lines = output.split('\n')
        self.assertEqual('configured_mail_sender configuration files:',
                         lines[0])

    def test_known_domains(self):
        sys.argv = ['test', '--list_domains', '--log_level=critical']
        capture_out = StringIO()
        self.saved_stdout = sys.stdout
        sys.stdout = capture_out
        with self.assertRaises(SystemExit) as e:
            main(email_sender=self.mock_sender)
        self.assertEqual(e.exception.code, 0,
                         "List domains shouldn't cause error")
        sys.stout = self.saved_stdout
        capture_out.seek(0)
        captured = capture_out.read()
        # lines = captured.split('\n')
        domains = {}
        for line in captured.split('\n')[1:]:
            line = line.replace(' ', '')
            if line:
                entry = line.split(':')
                key = entry[0]
                domains[key] = entry[1]
        self.assertEqual('smtp.mail.yahoo.com', domains.get('yahoo.com'))

    def test_initialize(self):
        os.removedirs(self.test_dir.name)
        with self.assertRaises(SystemExit) as e:
            self.runTest(extra_params=['--initialize'],
                         do_reference=False)
        self.assertEqual(0, e.exception.code)
        self.assertIsNone(self.mock_sender.result)
        # Should get same result
        files = os.listdir(self.test_dir.name)
        self.assertEqual(1, len(files))
        result = files[0]
        path = os.path.join(self.test_dir.name, result)
        with gzip.open(path, 'r') as f:
            saved_data = f.read()
        with open(TEST_CURRENT_DATA, 'r') as f:
            expected = bytes(f.read(), 'utf-8')
        self.assertEqual(expected, saved_data)

    def tearDown(self):
        if self.saved_stdout:
            sys.stdout = self.saved_stdout
        self.test_dir.cleanup()
        sys.argv = self.saved_argv



from io import StringIO
from typing import List, Tuple, Dict, Union

from configured_mail_sender import create_sender, MailSender
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from breeze_chms_api import breeze
from breeze_chms_api.profile_helper import ProfileHelper
from datetime import date, timedelta
from breeze_email_reports.table_format import ColSpec, _HTMLFormatter

import sys
import argparse
import os
import platformdirs
from breeze_email_reports._common import parse_arguments, check_date

APPLICATION_NAME = 'email_giver_report'
DEFAULT_LOG_FILE = os.path.join(platformdirs.user_log_dir(), APPLICATION_NAME + '.log')


def _field_name_to_id(helper: ProfileHelper) -> Dict[str, str]:
    """
    The Breeze profile data structure is way complex. Fields are generally identified
    by an id, but usually we just want a field by name.
    :param field_spec: Field specification from Breeze
    :param helper: Profile helper
    :return:
    """
    id_to_name = helper.get_field_id_to_name()
    return {v.split(':')[1] if ':' in v else v: k for (k,v) in id_to_name.items()}


def _get_date_range(begin: Union[str, None],
                    end: Union[str, None],
                    today: date = date.today()) -> Tuple[str, str]:
    """
    Generate date range from input parameters
    :param begin: Start date from command invocation
    :param end: End date from command invocation
    :param today: For testing purposes only, use as current day.
    :return: Interpreted beginning and ending dates
    All input and output dates are in the form YYYY-MM-DD.

    The logic is like this:
    if both dates given:
        Return dates as given
    else if end date (and only end date) given:
        return (beginning of hear from end date, end date)
    else if start date given:
        return (start date, end of start date month)
    else (neither was given)
        return first and last days of PREVIOUS month.
    """
    start_date = check_date(begin)
    end_date = check_date(end)

    # If neither start nor end given, use previous month
    if end_date:
        if not start_date:
            start_date = date(year=end_date.year, month=1, day=1)
        elif start_date > end_date:
            # Both given, but make sure it makes sense
            sys.exit("Start date can't be after end date")
        return start_date.isoformat(), end_date.isoformat()

    if start_date:
        # Start date given but no end
        if start_date > today:
            sys.exit("Start date can't be in the future")
        return start_date.isoformat(), today.isoformat()

    # If no dates given use beginning and end of the previous month.
    yr = today.year
    mo = today.month

    if mo == 1:
        start_date = date(yr - 1, 12, 1)
        end_date = date(yr - 1, 12, 31)
    else:
        start_date = date(yr, mo - 1, 1)
        end_date = (date(yr, mo, 1) - timedelta(days=1))
    return start_date.isoformat(), end_date.isoformat()


def get_contributions(breeze_api: breeze,
                      start_date: str,
                      end_date: str,
                      funds: List[str]) -> Tuple[Dict[str,
                                                      List[Tuple[str, float, str]]],
                                                 bool]:
    """
    Generate a contribution report:
    :param breeze_api: Breeze API to get contribution information
    :param start_date: Start of contribution range (YYY-MM-DD)
    :param end_date: End of contribution range
    :param funds: List of funds to include in report
    :return: (Dict profile's contributions, boolean True if there was any note.)
            Key in the dict is the profile id.
            Value for each profile is a list of contributions, each with:
                Date of contribution:
                Amount of contribution.
                The note.
    """

    all_funds = breeze_api.list_funds()
    name_to_id = {f['name']: f['id'] for f in all_funds}
    id_set = []
    for fn in funds:
        fid = name_to_id.get(fn)
        if not fid:
            sys.exit(f'Fund "{fn}" not found')
        id_set.append(fid)

    contributions = breeze_api.list_contributions(fund_ids=id_set,
                                                  start=start_date,
                                                  end=end_date)

    person_gifts = {}
    had_notes = False
    for contrib in contributions:
        amt = 0.0
        # Sum all contributions to requested funds within this contribution
        for fund in contrib['funds']:
            amt += float(fund['amount'])
        if amt > 0:     # Just checking, not sure how it could be otherwise
            paid_on = contrib['paid_on'][0:10]  # Just YYYY-MM-DD part
            pid = contrib.get('person_id')
            prv = person_gifts.get(pid, [])
            note = contrib.get('note')
            prv.append([paid_on, amt, note])
            if note:
                had_notes = True
            # Note: None is fine as a key. It will be interpreted as anonymous later.
            # person_gifts[pid] = prv
            person_gifts[pid] = prv

    ret_gifts = {pid: sorted(gifts, key=lambda x: x[0])
                 for pid, gifts in person_gifts.items()}

    return ret_gifts, had_notes


def main(breeze_api: breeze.BreezeApi = None,
         email_sender: MailSender = None):
    """
    Email a report of Breeze giving
    :param breeze_api: (For testing) pre-made Breeze API
    :param email_sender: (For testing) email sender
    :return: None
    """
    parser = argparse.ArgumentParser(APPLICATION_NAME)
    parser.add_argument('--from',
                        '-f',
                        dest='sender',
                        metavar='sender',
                        help='sending email address (required in most cases)')
    parser.add_argument('-s',
                        '--start',
                        metavar='StartDate',
                        help='Starting date as YYYY-MM-DD')
    parser.add_argument('-e',
                        '--end',
                        metavar='EndDate',
                        help='Ending date as YYYY-MM-D.')
    parser.add_argument('-t',
                        '--to',
                        metavar='Recipient(s)',
                        help='Public recipients')
    parser.add_argument('--cc',
                        metavar='CopyRecipients',
                        help='Copy recipients')
    parser.add_argument('--bcc',
                        metavar='BlindRecipients',
                        help='Blind recipients')
    parser.add_argument('--totals',
                        default=False,
                        action='store_const',
                        const=True,
                        help='Include total for each giver')
    parser.add_argument('--summary',
                        default=False,
                        action='store_const',
                        const=True,
                        help='Show only total'
                        )
    parser.add_argument('--breeze_creds',
                        metavar='<Breeze credential file>',
                        help='file with Breeze credentials if not standard',
                        default=None)
    parser.add_argument('--email_creds',
                        metavar='<email credential file>',
                        help='file with email passwords if not standard')
    parser.add_argument('--email_servers',
                        metavar='<email domain specification file>',
                        help='file with email domain configuration if not standard')

    args = parse_arguments(parser, APPLICATION_NAME,
                           [(['funds'],
                             {'nargs': '*',
                              'metavar': 'Fund(s)',
                              'help': 'Fund(s) to include'})])

    if not args.sender:
        sys.exit('--from is required')
    if not (args.to or args.cc or args.bcc):
        sys.exit('At least one of --to, --cc or --bcc is required')

    if args.summary:
        args.total = False

    if not args.funds:
        sys.exit('One or more funds is required')

    start_date, end_date = _get_date_range(args.start, args.end)

    if not breeze_api:
        # Normal case not testing
        breeze_api = breeze.breeze_api(overrides=args.breeze_creds)

    contributions, have_notes = get_contributions(breeze_api,
                                                  start_date=start_date,
                                                  end_date=end_date,
                                                  funds=args.funds)

    profile_fields = breeze_api.get_profile_fields()
    profile_helper = ProfileHelper(profile_fields)
    field_name_to_id = _field_name_to_id(profile_helper)
    name_id = field_name_to_id.get("Name")
    address_id = field_name_to_id.get("Address")
    phone_id = field_name_to_id.get("Phone")
    email_id = field_name_to_id.get("Email")
    id_fields = [name_id, address_id, phone_id, email_id]

    needed_profile_ids = [id for id in contributions.keys() if id]
    # Note: I'd use json_filter here to just get the interesting ids, but
    # there's no documentation on how to do that. And doing separate calls
    # for each profile would risk getting throttled. So we'll just
    # get everything and pick the ones we want.
    # Instead, if there are only a few profiles, pull them individually.
    # Otherwise, just get the whole lot in one request to avoid throttling.
    if len(needed_profile_ids) > 10:
        # too many people so get all at once
        profiles = breeze_api.list_people(details=True)
        profiles = [p for p in profiles if p['id'] in needed_profile_ids]
    else:
        # Just get the few we need
        profiles = [breeze_api.get_person_details(id) for id in needed_profile_ids]

    processed = {p['id']: profile_helper.process_member_profile(p)
                 for p in profiles if p['id'] in needed_profile_ids}

    toformat = []
    grand_total = 0.0
    for id, p in contributions.items():
        if id:
            details = processed.get(id)
            id_data = []
            for fld in id_fields:
                dat = details.get(fld)
                if dat:
                    if isinstance(dat, List):
                        dat = '\n'.join(dat)
                    dat = dat.replace(';', '\n')
                    id_data.append(dat)
        else:
            id_data = ['Anonymous']
        # Fix numeric values and get total.
        total = 0.0
        for row in p:
            amt = row[1]
            total += amt
            row[1] = f'{amt:,.2f}'
        col1 = '\n'.join(id_data)
        grand_total += total
        if args.summary:
            toformat.append([col1, [[f'{total:,.2f}']]])
        else:
            if len(p) > 1 and args.totals:
                p.append(['Total', f'{total:,.2f}', ''])
            toformat.append([col1, [list(v) for v in p]])

    # So toformat now is a list of all givers. Each item in the list
    # is a list (why not a Tuple?) where the first entry is the info
    # about the giver (name, address...) as newline-separated string,
    # and following is a list with all the contribution items, the
    # interesting bits being date, amount, notes.

    # sort by giver name (last, first)
    toformat = sorted(toformat, key=lambda x: x[0])
    if grand_total > 0:
        if args.summary:
            toformat.append(['Grand Total', [[f'{grand_total:,.2f}']]])
        else:
            toformat.append(['Grand Total', [['All', f'{grand_total:,.2f}', '']]])

    # Now, let's set up columns.
    # Columns will be:
    #   Donor
    #   Date (unless only totals selected)
    #   Amount
    #   Notes (if there are notes)
    columns = [ColSpec('Donor', width=15, header_attrs={'text-align': 'start'})]
    # args.summary means we only want total for each giver, so no dates.
    if not args.summary:
        columns.append(ColSpec('Date', width=10,
                               header_attrs={'text-align': 'center'}))
    columns.append(ColSpec('Amount', width=7, header_attrs={'text-align': 'end'}))
    # Only create a Notes column if there actually are notes.
    if not args.summary and have_notes:
        columns.append(ColSpec('Note'))

    msg = MIMEMultipart()
    fnames = ', '.join(args.funds)
    subject = f'Contributions to fund(s)\n{fnames}\nfrom {start_date} to {end_date}'
    msg['Subject'] = subject.replace('\n', ' ')
    if args.to:
        msg['to'] = args.to
    if args.bcc:
        msg['bcc'] = args.bcc
    if args.cc:
        msg['cc'] = args.cc
    if toformat:
        # msg.attach(MIMEText(f'{subject}\n\n'))
        # As of now, only HTML format is supported. Maybe a future version.
        formatter = _HTMLFormatter(column_specs=columns, id_top=False, caption=subject)
        output = StringIO()
        formatter.format_table(toformat, output)
        output.seek(0)
        txt = output.read()
        # with open('test.html', 'w') as f:
        #     f.write(txt)
        msg.attach(MIMEText(txt, 'html'))
    else:
        msg.attach(MIMEText(f'No contributions found'))

    sender = email_sender if email_sender else \
        create_sender(args.sender,
                      creds_file=args.email_creds,
                      overrides=args.email_servers)
    sender.send_message(msg)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Tuple, Dict, Union

from breeze_chms_api import breeze
from configured_mail_sender import known_domains
from datetime import date, timedelta

import logging
import configured_mail_sender
import argparse
import sys
import os
import platformdirs


def _list_domains(args):
    domains = known_domains(overrides=args.email_servers)
    print("Known email domains:")
    for domain, text in domains.items():
        print(f'   {domain}: {text}')


def parse_arguments(parser: argparse.ArgumentParser,
                    application_name: str,
                    to_add: List[Tuple[List[str], Dict[str, str]]]=[]) -> argparse.Namespace:
    """
    Fill in argument parser with special debug options
    :param parser: The parser
    :param application_name: Name of application
    :param to_add: List of argument specs for additional parameters
    :return: None
    """

    logfile = os.path.join(platformdirs.user_log_dir(), f'{application_name}.log')

    parser.add_argument('--list_directories',
                        default=False,
                        action='store_true',
                        help='list data and configuration directories and exit')
    parser.add_argument('--list_domains',
                        default=False,
                        action='store_true',
                        help='Print known email domains and exit')
    parser.add_argument('--logfile',
                        default=logfile,
                        metavar='Log file',
                        help='File for application logs')
    parser.add_argument('--log_level',
                        default='info',
                        metavar='Log level',
                        help='Logging level, default=info')

    for parm in to_add:
        parser.add_argument(*parm[0], **parm[1])

    args = parser.parse_args()

    if args.logfile == 'stdout':
        out = {'stream': sys.stdout}
    elif args.logfile == 'stderr':
        out = {'stream': sys.stderr}
    else:
        out = {'filename': args.logfile}
        directory, file = os.path.split(args.logfile)
        Path(directory).mkdir(exist_ok=True, parents=True)

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=getattr(logging, args.log_level.upper()),
        **out
    )
    logging.info(f'Running {" ".join(sys.argv)}')

    if args.list_domains:
        _list_domains(args)
        sys.exit(0)

    if args.list_directories:
        # Making it easy to note where users should look for files on this platform.
        mailsender_configs = configured_mail_sender.config_file_list()
        print('configured_mail_sender configuration files:')
        for f in mailsender_configs:
            print(f'\t{f}')
        breeze_config = breeze.config_file_list(overrides=args.breeze_creds)
        print('breeze_chms_api configuration files:')
        for f in breeze_config:
            print(f'\t{f}')
        if hasattr(args, 'data'):
            print(f'{application_name} data directory:\n\t{args.data}')
        print(f'log output\n\t{args.logfile}')
        sys.exit(0)

    return args


def check_date(dt: str) -> Union[str, None]:
    """
    Check for correctly formatted date and convert to date object
    Exit with error if date format is bad.
    :param dt:
    :return: Date as date object, or None if none.
    """
    try:
        return date.fromisoformat(dt) if dt else None
    except ValueError as e:
        # There was a problem with the date
        sys.exit(f'"{dt}" is not a valid date. Needs to be YYYY-MM-DD')



import csv
import html
from abc import abstractmethod
from io import IOBase
from typing import List, Tuple, Union, Dict

# Pad string to length, e.g. f'{a<15}' will pad to 15 chars.


class ColSpec:
    def __init__(self,
                 title: str,
                 header_attrs: Dict[str, str] = {},
                 row_attrs: Dict[str, str] = {},
                 width: int = 0):
        """
        Column specifications for table formatter
        :param title: Column title
        :param header_attrs: HTML attributes for header (if appropriate)
        :param row_attrs: HTML attributes for data rows (if appropriate)
        :param width: width of column for text formatting
        """
        self.title_val = title
        self.header_attrs_val = header_attrs if header_attrs else row_attrs
        self.row_attrs_val = row_attrs if row_attrs else header_attrs
        self.width_val = width

    @property
    def title(self):
        return self.title_val

    @property
    def header_attrs(self):
        return self.header_attrs_val

    @property
    def row_attrs(self):
        return self.row_attrs_val

    @property
    def width(self):
        return self.width_val


class _ReportFormatter:
    def __init__(self,
                 column_specs: List[ColSpec],
                 caption: str = "",
                 id_top: bool = True):
        """
        Base settings for report formatters
        :param column_specs: List of column specifications
        :param caption: Caption to put at top of table
        :param id_top: If true(default) identification information is in a row at
            the top of each group. If false, It's in its own spanned column to
            the left of the data.
        """
        self.column_specs = column_specs
        self.columns = len(column_specs)
        self.id_top = id_top
        self.caption = caption

    @abstractmethod
    def format_table(self,
                     data: List[Tuple[str, List[Union[str, List[str]]]]],
                     output: IOBase):
        """
        Format the data in appropriate format and write to output
        :param data: Data for the table
        :param output: Output stream to write to.
        :return: None
        The input data is a list of tuples. The first value is a group header
        that divides subsequent data. (A member name in this context.) The second
        value is a list of tuples, each with label for the tuple followed by values
        for columns in the row. Each column value can be None, a string, or a list
        of strings.

        Note: If self.id_top is true, the string in the first value of each Tuple
        should go in a spanned row the data entry rows.
        If self.id_top is false, the data in the first value of each tuple goes
        in the first column of each group, spanned down across the data values.
        The information from the list of values goes in subsequent rows.
        """
        pass


def _tokenize(v):
    if isinstance(v, float):
        return [f'{v:.2f}']
    elif isinstance(v, str):
        return v.split(';')
    else:
        return v


def _escape_html_value(val: Union[str, List[str]]) -> str:
    if val:
        if isinstance(val, list):
            # For lists in HTML break each item on semicolons (for addresses).
            # But we have to do that before the html.escape() because that
            # introduces semicolons.
            return '<br/>'.join([html.escape(i) for elt in val
                                 for i in _tokenize(elt)
                                 ])
        else:
            return html.escape(val)
    else:
        return ' '


def _attr_str(attrs: Dict[str, str]) -> str:
    """
    Return a string suitable for html style of attribute map
    :param attrs: Attributes
    :return: Attributes turned into a string
    """
    return ' '.join([f'{k}: {v};' for k, v in attrs.items()])


class _HTMLFormatter(_ReportFormatter):
    def __init__(self,
                 *args,
                 top_attrs: Dict[str, str] = {},
                 **kwargs):
        """
        Format as HTML
        :param args: Standard arguments
        :param top_attrs: if id_top is set, HTML attributes for the spanned id row
        :param kwargs:
        """
        _ReportFormatter.__init__(self, *args, **kwargs)
        self.table_header = ('<table border="1" style="border:1px solid #000000;'
                             'border-collapse:collapse" cellpadding="4">\n')
        self.top_attrs = top_attrs

    def format_table(self,
                     data: List[Tuple],
                     output: IOBase):
        """
        Format as an HTML table
        :param data:
        :param output:
        :return:
        """
        output.write('<head>\n<style>\n')
        for c in range(len(self.column_specs)):
            output.write(f'#r1 td:nth-child({c+1}) '
                         f'{{ {_attr_str(self.column_specs[c].row_attrs)} }}\n')
            output.write(f'th td:nth-child({c+1}) '
                         f'{{ {_attr_str(self.column_specs[c].header_attrs)} }}\n')
        for c in range(1, len(self.column_specs)):
            output.write(f'#rn td:nth-child({c}) '
                         f'{{ {_attr_str(self.column_specs[c].row_attrs) }}}\n')

        if self.id_top:
            output.write(f'#hr {{ {_attr_str(self.top_attrs)} }}\n')

        column_specs = self.column_specs
        if not self.id_top:
            # Ignore first column when outputting data rows
            column_specs = column_specs[1:]
        number_columns = len(column_specs)

        output.write(f'th {{ {_attr_str(self.column_specs[0].header_attrs)} }}\n')
        output.write('</style>\n</head>\n<body>\n')

        output.write('<table border="1" style="border:1px thick double black";'
                     'border-collapse:collapse" cellpadding="4">\n')
        if self.caption:
            cap = self.caption.replace('\n', '<br/>')
            output.write(f'<caption><b>{cap}</b></caption>\n')
        nl = '\n'
        row = nl.join([f'  <th>{c.title}</th>'
                       for c in self.column_specs])
        output.write(f'{nl} <tr>{nl}{row}{nl} </tr>{nl}')

        for name, values in data:
            name = html.escape(name)
            if self.id_top:
                # Identifying info goes in first row at the top
                output.write(f'<tr id="hr">\n  <td colspan="{self.columns}" '
                             f'<b>{name}</b></td>\n</tr>\n')
                first_col_data = None
            else:
                # Identifying information goes in first column of first row
                person_info = name.replace('\n', '<br/>\n')
                first_col_data = (f'<tr id="r1"><td rowspan={len(values)}>\n'
                                  f'{person_info}</td>\n')

            for row in values:
                if first_col_data:
                    output.write(first_col_data)
                    first_col_data = None
                else:
                    output.write(f' <tr id="rn">\n')
                for i in range(number_columns):
                    # cs = column_specs[i]
                    colval = _escape_html_value(row[i])
                    output.write(f'  <td>{colval}</td>\n')
                output.write(' </tr>\n')

        output.write('</table>\n')


def _splitval(val: str, cs: ColSpec, splits: str = ';: ') \
        -> Tuple[str, Union[str, None]]:
    """
    Split a string if too long
    :param val: Value to split
    :param cs: ColSpec with desired length
    :return: Tuple, parts of val before and after the split
    """
    # use .strip() to remove pre and post whitespace
    if not val:
        return '', None
    if len(val) <= cs.width:
        return val, ''
    last_found = cs.width - 1
    found_char = None
    for c in splits:
        end = cs.width if c == ' ' else cs.width - 1
        found = val.rfind(c, 0, end)
        if found > 0:
            last_found = found
            found_char = c
            break

    first = val[:last_found] if found_char == ' ' else val[:last_found + 1]
    second = val[last_found + 1:]
    return first, second


class _ToFileFormatter(_ReportFormatter):
    def __init__(self, *args, **kwargs):
        _ReportFormatter.__init__(self, *args, **kwargs)
        # self.line_mark = '   |'
        # self.row_prefix = self.line_mark
        # for cs in self.column_specs:
        #     self.line_mark += ''.rjust(cs.width, '-') + '|'

    def format_table(self,
                     data: List[Tuple],
                     output: IOBase):
        """
        Format as an text table
        :param data:
        :param output:
        :return:
        """
        for section in data:
            name, values = section
            self.start_section(name, output)
            for row in values:
                more = True
                self.start_group(output)
                # print(self.line_mark, file=output)
                while more:
                    # print(self.row_prefix, end='', file=output)
                    more = False
                    next_values = []
                    this_row = []
                    for i in range(min(len(row), self.columns)):
                        cs = self.column_specs[i]
                        colval = row[i]
                        nextval = []
                        if colval:
                            if isinstance(colval, list):
                                curval = colval[0]
                                nextval = colval[1:]
                                if nextval:
                                    more = True
                            else:
                                curval = colval
                                nextval = []
                            curval, overflow = self.splitval(curval, cs)
                            if overflow:
                                nextval = [overflow] + nextval
                                more = True
                        else:
                            curval = ''
                        this_row.append(curval if curval else '')
                        next_values.append(nextval)

                    self.emit_row(this_row, output)
                    row = next_values

            # print(self.line_mark, file=output)
            self.end_section(output)
        self.finish_output(output)

    @abstractmethod
    def start_section(self, name: str, output: IOBase):
        pass

    def end_section(self, output: IOBase):
        pass

    def start_group(self, output):
        pass

    def splitval(self, val: str, cs: ColSpec) -> Tuple[str, Union[str, None]]:
        return val, None

    def emit_row(self, values: List[str], output: IOBase):
        pass

    def finish_output(self, output: IOBase):
        pass


class _TextFormatter(_ToFileFormatter):
    def __init__(self, *args, **kwargs):
        _ToFileFormatter.__init__(self, *args, **kwargs)
        self.line_mark = '   |'
        self.row_prefix = self.line_mark
        for cs in self.column_specs:
            self.line_mark += ''.rjust(cs.width, '-') + '|'

    def start_section(self, name: str, output: IOBase):
        print('', file=output)
        print(name, file=output)

    def start_group(self, output):
        print(self.line_mark, file=output)

    def splitval(self, val: str, cs: ColSpec) -> Tuple[str, str]:
        ret_val, overflow = _splitval(val, cs)
        return ret_val.replace('\n', ';'), ('+' + overflow) if overflow else None

    def emit_row(self, values: List[str], output: IOBase):
        print(self.row_prefix, end='', file=output)
        for i in range(len(values)):
            v = values[i]
            cs = self.column_specs[i]
            v = v.ljust(cs.width)
            print(v + '|',
                  end='',
                  file=output)
            output.flush()
        print('', file=output)

    def end_section(self, output: IOBase):
        print(self.line_mark, file=output)


class _CSVFormatter(_ToFileFormatter):
    def __init__(self, *args, **kwargs):
        _ToFileFormatter.__init__(self, *args, **kwargs)
        self.writer = None

    def start_section(self, name: str, output: IOBase):
        if not self.writer:
            self.writer = csv.writer(output)
        self.writer.writerow([name])

    def emit_row(self, values: List[str], output: IOBase):
        self.writer.writerow([''] + values)



import json
import os
import unittest
import tempfile
from datetime import date
from typing import Tuple, List, Union
from breeze_email_reports._common import check_date
import sys
from breeze_email_reports.EmailGiverReport import main, _get_date_range

"""
Some notes on the test data...
Profiles from CurrentDataBreeze.json:
    13701083: Firstname1 Alast
    13711803: Firstname2 Blast
    19870634: NewFirst Bonzo
    13857066: Firstname3 Blast

Funds in funds.json:
    1816906: Budget (was 2024 budget)
    1816907: Budget (Non Exempt) (Was 2024...)
    1901063: Memorial (Was Anne Cook Memorial)
    1520829: Rents
"""

# TODO: Verify Breeze API calls. In particular requested funds and dates
# TODO: Test logged output (stdout and stdin?)

TEST_FILES_DIR = os.path.join(os.path.split(__file__)[0], 'test_files')


# Returned by Breeze for "current" profile data
CURRENT_DATA_BREEZE = os.path.join(TEST_FILES_DIR, 'CurrentDataBreeze.json')
FUNDS = os.path.join(TEST_FILES_DIR, 'GivingFunds.json')
GIFTS = os.path.join(TEST_FILES_DIR, 'Gifts.json')
EXTRA_GIFTS = os.path.join(TEST_FILES_DIR, 'GiftsExtra.json')
EXTRA_PROFILES = os.path.join(TEST_FILES_DIR, 'ExtraProfiles.json')

# Expected results in email
EXPECTED_HTML = os.path.join(TEST_FILES_DIR, 'GiversExpectedHTML.html')
EXPECTED_HTML_MORE = os.path.join(TEST_FILES_DIR, 'GiversExpectedHTML.more.html')
EXPECTED_SUMMARY = os.path.join(TEST_FILES_DIR, 'GiversExpectedSummary.html')

TO_ADDRESS = 'to@test.com'
BCC_ADDRESS = 'bcc1@bcc.com, bcc2@bcc.com'
CC_ADDRESS = 'cc@cc.com'


class MockBreezeAPI:
    def __init__(self, fields: str,
                 profiles: str,
                 funds: List[str] = [], contributions: List[str] = []):
        self.fields = fields
        self.profiles = profiles
        self.funds = funds
        self.contributions = contributions
        self.funds_requested = None
        self.start_requested = None
        self.end_requested = None
        self.profile_map = None
        self.list_people_requests = []
        self.person_detail_requests = []

    def list_people(self, **kwargs):
        self.list_people_requests.append(kwargs)
        return self.profiles

    def get_profile_fields(self):
        return self.fields

    def get_person_details(self, person_id):
        if not self.profile_map:
            self.profile_map = {p.get("id"): p for p in self.profiles}
        self.person_detail_requests.append(person_id)
        return self.profile_map.get(person_id)

    def list_funds(self):
        ret = []
        for fundfile in self.funds:
            with open(fundfile, 'r') as f:
                ret += json.loads(f.read())
        return ret

    def list_contributions(self,
                           fund_ids: List[str] = [],
                           start: str = None,
                           end: str = None):
        self.funds_requested = fund_ids
        self.start_requested = start
        self.end_requested = end
        ret = []
        for contributions in self.contributions:
            with open(contributions, 'r') as f:
                ret += json.loads(f.read())
        return ret


class MockSender:
    def __init__(self):
        self.result = None

    def send_message(self, msg):
        self.result = msg


class TestGiverSender(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.mock_sender = MockSender()
        self.saved_argv = sys.argv
        self.saved_stdout = None
        self.mock_api = None

    def make_api(self, fields=[], profiles=[], funds=[FUNDS], contributions=[]):
        self.mock_api = MockBreezeAPI(fields, profiles, funds=funds,
                                      contributions=contributions)

    def runTest(self,
                # report_format: str = 'text',
                contributions: List[str] = [GIFTS],
                extra_profiles: List[str] = [],
                start_date: str = '2024-10-01',
                end_date: str = '2024-10-31',
                extra_params: List[str] = []) -> Tuple[str, int]:
        """
        Run BreezeProfile report. Return emitted event.
        :param contributions: file with contributions to return
        :param extra_profiles: More profiles to load
        :param start_date: first date
        :param end_date: end date
        :param extra_params:
        :return:
        :raises: ParamException, IOException
        """

        with open(CURRENT_DATA_BREEZE, 'r') as f:
            fields, profiles = json.loads(f.read())

        for extra in extra_profiles:
            with open(extra, 'r') as f:
                profiles += (json.loads(f.read()))

        self.make_api(fields=fields, profiles=profiles, funds=[FUNDS],
                      contributions=contributions)

        sys.argv = ['test', '-f', 'from@test.com', '--to', TO_ADDRESS,
                    '--bcc', BCC_ADDRESS,
                    '--cc', CC_ADDRESS,
                    # '--data', self.test_dir.name,
                    '--log_level=critical',
                    f'--logfile={os.path.join(self.test_dir.name, "test.log")}']
        # Including format TBD
        if start_date:
            sys.argv += ['--start', start_date]
        if end_date:
            sys.argv += ['--end', end_date]

        sys.argv += extra_params

        ret = main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual(None, ret)

    def test_basic(self):
        self.runTest(extra_params=['--totals', 'Budget'])
        sent_message = self.mock_sender.result
        # self.assertEqual('Breeze profile change report',
        #                      str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('to'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('bcc'))
        self.assertEqual(CC_ADDRESS, sent_message.get('cc'))

        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        # pl1 = payloads[1]
        # self.assertEqual(pl0.get_content_subtype(), 'plain')
        # Contributions to fund(s)
        # Budget
        # from 2024-10-01 to 2024-10-31
        # self.assertTrue(str(pl0.get_payload()).
        #                 startswith('Breeze changes for Unknown thru'))
        self.assertEqual(pl0.get_content_subtype(), 'html')
        got = str(pl0.get_payload())
        # with open(EXPECTED_HTML, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_HTML, 'r') as f:
            self.assertEqual(got, f.read())

    def test_summary(self):
        self.runTest(extra_params=['--summary', 'Budget'])
        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        got = str(pl0.get_payload())
        # with open(EXPECTED_SUMMARY, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_SUMMARY, 'r') as f:
            self.assertEqual(got, f.read())

    def test_more(self):
        self.runTest(extra_params=['Budget'],
                     contributions=[GIFTS, EXTRA_GIFTS],
                     extra_profiles=[EXTRA_PROFILES])
        sent_message = self.mock_sender.result
        # self.assertEqual('Breeze profile change report',
        #                      str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('to'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('bcc'))
        self.assertEqual(CC_ADDRESS, sent_message.get('cc'))

        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        # self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertEqual(pl0.get_content_subtype(), 'html')
        got = str(pl0.get_payload())
        # with open(EXPECTED_HTML_MORE, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_HTML_MORE, 'r') as f:
            self.assertEqual(f.read(), got)
        # Contributions to fund(s) (pl0.get_payload())
        # Budget
        # from 2024-10-01 to 2024-10-31
        # self.assertTrue(str(pl0.get_payload()).
        #                 startswith('Breeze changes for Unknown thru'))

    def test_no_data(self):
        self.runTest(extra_params=['Budget'],
                     contributions=[])
        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        self.assertEqual("No contributions found", payloads[0].get_payload())

    def test_no_funds(self):
        sys.argv = ['test', '--bcc', 'foo@bar', '-f', 'foo']
        self.make_api()
        with self.assertRaises(SystemExit) as se:
            main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual('One or more funds is required', se.exception.code)

    def test_bad_fund(self):
        self.make_api()
        badfund = 'nofund'
        sys.argv = ['test', '--bcc', 'foo@bar', '--from', 'test@foo.com', badfund]
        with self.assertRaises(SystemExit) as se:
            main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual(f'Fund "{badfund}" not found', se.exception.code)

    def test_no_sender(self):
        self.make_api()
        sys.argv = ['test', '--bcc', 'foo@bar', 'afund']
        with self.assertRaises(SystemExit) as se:
            main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual('--from is required', se.exception.code)

    def test_no_receiver(self):
        self.make_api()
        sys.argv = ['test', '--from', 'too@bar', 'afund']
        with self.assertRaises(SystemExit) as se:
            main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual('At least one of --to, --cc or --bcc is required',
                         se.exception.code)

    def test_no_contributions(self):
        self.make_api()
        start_date = '2024-10-01'
        end_date = '2024-11-15'
        self.runTest(extra_params=[
                                   '--start', start_date,
                                   '--end', end_date,
                                   'Budget'],
                     contributions=[],
                     extra_profiles=[EXTRA_PROFILES])
        # Note: The start and end dates don't actually impact the returned
        # data in the test. "We assume Breeze filters correctly given the
        # supplied parameters. We do make sure that the tool sends the
        # appropriate query.
        sent_message = self.mock_sender.result

        self.assertEqual(TO_ADDRESS, sent_message.get('to'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('bcc'))
        self.assertEqual(CC_ADDRESS, sent_message.get('cc'))

        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertEqual(pl0.get_payload(), 'No contributions found')

    def test_list_dir(self):
        sys.argv = ['test', '--list_directories']
        with self.assertRaises(SystemExit) as se:
            main(breeze_api=self.mock_api, email_sender=self.mock_sender)
        self.assertEqual(se.exception.code, 0)

    def date_range(self,
                   begin: Union[None, str],
                   end: Union[None, str],
                   expect_start: str,
                   expect_end: str,
                   today: str = '2024-11-09'):
        """
        Use case for testing date ranges, All dates are in form YYYY-MM-DD
        :param begin: Testing begin date
        :param end: Testing end date
        :param expect_start: Expected output start
        :param expect_end: Expected output end
        :param today: Use as "today" for testing purposes
        :return: None
        """
        result_start, result_end = _get_date_range(begin, end,
                                                   today=date.fromisoformat(today))
        self.assertEqual(expect_start, result_start)
        self.assertEqual(expect_end, result_end)

    def test_date_range(self):
        # Bad date causes exit.
        bad = '2024-02-30'
        with self.assertRaises(SystemExit) as e:
            check_date(bad)
        self.assertEqual(f'"{bad}" is not a valid date. Needs to be YYYY-MM-DD',
                         e.exception.code)

        # If no dates, return previous month. (Leap year.)
        self.date_range(None,
                        None,
                        '2024-02-01',
                        '2024-02-29',
                        '2024-03-15')

        # If no dates, return previous month. (Previous year.)
        self.date_range(None,
                        None,
                        '2023-12-01',
                        '2023-12-31',
                        '2024-01-15')

        # Only begin date, return begin date until today
        self.date_range('2023-12-01',
                        None,
                        '2023-12-01',
                        '2024-01-15',
                        '2024-01-15')

        # Only end date, return start of year until end date
        self.date_range(None,
                        '2023-12-05',
                        '2023-01-01',
                        '2023-12-05',
                        '2024-01-15')

        # Explicit beginning and end return same
        self.date_range('2023-09-23',
                        '2023-12-05',
                        '2023-09-23',
                        '2023-12-05',
                        '2024-01-15')

        # Explicit dates inconsistent exits with error
        with self.assertRaises(SystemExit) as e:
            self.date_range('2023-12-23',
                            '2023-01-05',
                            '2023-09-23',
                            '2023-12-05',
                            '2024-01-15')
        self.assertEqual("Start date can't be after end date", e.exception.code)

        # Start date in future
        with self.assertRaises(SystemExit) as e:
            self.date_range('2025-01-01',
                            None,
                            '',
                            '',
                            '2024-01-01')
        self.assertEqual("Start date can't be in the future", e.exception.code)

    def tearDown(self):
        if self.saved_stdout:
            sys.stdout = self.saved_stdout
        self.test_dir.cleanup()
        sys.argv = self.saved_argv
